""" Entry point for Next-action's command-line interface. """

import argparse
import os

from next_action.arguments import parse_arguments
from next_action.pick_action import next_actions
from next_action.todotxt import read_todotxt_files, Tasks
from next_action.output import render


def validate_arguments(namespace: argparse.Namespace, tasks: Tasks) -> str:
    """ Check whether the context and projects given on the command line actually exist in the task file. """
    unknown_contexts = (namespace.contexts | namespace.excluded_contexts) - tasks.contexts()
    unknown_projects = (namespace.projects | namespace.excluded_projects) - tasks.projects()
    error_messages = []
    if unknown_contexts:
        plural = "s" if len(unknown_contexts) > 1 else ""
        error_messages.append("unknown context{0}: {1}".format(plural, ", ".join(sorted(unknown_contexts))))
    if unknown_projects:
        plural = "s" if len(unknown_projects) > 1 else ""
        error_messages.append("unknown project{0}: {1}".format(plural, ", ".join(sorted(unknown_projects))))
    return (" (warning: {0})".format("; ".join(error_messages))) if error_messages else ""


def next_action() -> None:
    """ Entry point for the command-line interface.

        Basic recipe:
        1) parse command-line arguments,
        2) read todo.txt file(s),
        3) determine the next action(s) and,
        4) display them.
    """
    parser, namespace = parse_arguments()
    filenames = [os.path.expanduser(filename) for filename in namespace.file]
    try:
        tasks = read_todotxt_files(filenames)
    except OSError as reason:
        parser.error("can't open file: {0}".format(reason))
    actions = next_actions(tasks, namespace)
    print(render(actions, namespace) if actions else "Nothing to do!" + validate_arguments(namespace, tasks))



""" Unit tests for the color module in the output package. """

import unittest

from next_action.output import colorize


class ColorizeTest(unittest.TestCase):
    """ Unit tests for the colorize method. """

    def test_no_style(self):
        """ Test that the output is unchanged when no style is passed. """
        task = "(A) Task @home +Project"
        self.assertEqual(task, colorize(task, ""))

    def test_default_style(self):
        """ Test that the output is styled when a style is passed. """
        task = "(A) Task @home +Project"
        self.assertEqual('\x1b[38;5;18;01m(A)\x1b[39;00m Task \x1b[38;5;124m@home\x1b[39m \x1b[38;5;9m+Project\x1b[39m',
                         colorize(task, "default"))

    def test_strip_added_linebreak(self):
        """ Test that the linebreak added by Pygments is removed. """
        task = "(A) Task @home +Project\n"
        self.assertEqual(
            '\x1b[38;5;18;01m(A)\x1b[39;00m Task \x1b[38;5;124m@home\x1b[39m \x1b[38;5;9m+Project\x1b[39m\n',
            colorize(task, "default"))



""" Collection of Todo.txt tasks. """

from typing import List, Set

from .task import Task


class Tasks(List[Task]):
    """ Collection of Todo.txt tasks. """

    def contexts(self) -> Set[str]:
        """ Return the contexts used in the collection of tasks. """
        return set([context for task in self for context in task.contexts()])

    def projects(self) -> Set[str]:
        """ Return the projects used in the collection of tasks. """
        return set([project for task in self for project in task.projects()])



""" Unit tests for the configration file parsing. """

import os
import sys
import unittest
from unittest.mock import patch, call, mock_open

import yaml

from next_action.arguments import config, parse_arguments

from .test_parser import USAGE_MESSAGE


class ConfigTestCase(unittest.TestCase):
    """ Base class for configuration file unit tests. """
    def setUp(self):
        os.environ['COLUMNS'] = "120"  # Fake that the terminal is wide enough.


class ReadConfigFileTest(ConfigTestCase):
    """ Unit tests for reading the config file. """

    @patch.object(sys, "argv", ["next-action", "--config-file", "config.cfg"])
    @patch.object(config, "open", mock_open(read_data=""))
    def test_empty_file(self):
        """ Test that an empty config file doesn't change the filenames. """
        self.assertEqual(["~/todo.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open")
    def test_missing_default_config(self, mock_file_open):
        """ Test that a missing config file at the default location is no problem. """
        mock_file_open.side_effect = FileNotFoundError("some problem")
        self.assertEqual(["~/todo.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="- this_is_invalid"))
    @patch.object(sys.stderr, "write")
    def test_invalid_document(self, mock_stderr_write):
        """ Test that a config file that's not valid YAML raises an error. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: ~/.next-action.cfg is invalid: '['this_is_invalid']' is not a "
                               "document, must be a dict\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="foo: bar"))
    @patch.object(sys.stderr, "write")
    def test_no_file_key(self, mock_stderr_write):
        """ Test that a config file without file key doesn't change the filenames. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE),
                          call("next-action: error: ~/.next-action.cfg is invalid: foo: unknown field\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--config-file", "config.cfg"])
    @patch.object(config, "open")
    @patch.object(sys.stderr, "write")
    def test_error_opening(self, mock_stderr_write, mock_file_open):
        """ Test a config file that throws an error. """
        mock_file_open.side_effect = OSError("some problem")
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE), call("next-action: error: can't open file: some problem\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--config-file", "config.cfg"])
    @patch.object(config, "open")
    @patch.object(sys.stderr, "write")
    def test_error_parsing(self, mock_stderr_write, mock_file_open):
        """ Test a config file that throws a parsing error. """
        mock_file_open.side_effect = yaml.YAMLError("some problem")
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE), call("next-action: error: can't parse config.cfg: some problem\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--config-file", "config.cfg"])
    @patch.object(config, "open")
    @patch.object(sys.stderr, "write")
    def test_file_not_found(self, mock_stderr_write, mock_file_open):
        """ Test a config file that throws an error. """
        mock_file_open.side_effect = FileNotFoundError("some problem")
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual([call(USAGE_MESSAGE), call("next-action: error: can't open file: some problem\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--config-file"])
    @patch.object(config, "open")
    def test_skip_config(self, mock_file_open):
        """ Test that the config file is not read if the user doesn't want to. """
        parse_arguments()
        self.assertEqual([], mock_file_open.call_args_list)


class WriteConfigFileTest(ConfigTestCase):
    """ Unit tests for the write-config-file argument. """

    @patch.object(sys, "argv", ["next-action", "--write-config-file"])
    @patch.object(config, "open", mock_open(read_data=""))
    @patch.object(sys.stdout, "write")
    def test_default_file(self, mock_stdout_write):
        """ Test that a config file can be written to stdout. """
        self.assertRaises(SystemExit, parse_arguments)
        expected = "# Configuration file for Next-action. Edit the settings below as you like.\n"
        expected += "file: ~/todo.txt\nnumber: 1\nreference: multiple\nstyle: default\n"
        self.assertEqual([call(expected)], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--write-config-file", "--number", "3", "--reference", "never",
                                "--style", "fruity", "--file", "~/tasks.txt"])
    @patch.object(config, "open", mock_open(read_data=""))
    @patch.object(sys.stdout, "write")
    def test_with_args(self, mock_stdout_write):
        """ Test that the config file written to stdout includes the other arguments. """
        self.assertRaises(SystemExit, parse_arguments)
        expected = "# Configuration file for Next-action. Edit the settings below as you like.\n"
        expected += "file: ~/tasks.txt\nnumber: 3\nreference: never\nstyle: fruity\n"
        self.assertEqual([call(expected)], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--write-config-file", "--file", "~/tasks.txt", "--file", "project.txt"])
    @patch.object(config, "open", mock_open(read_data=""))
    @patch.object(sys.stdout, "write")
    def test_multiple_files(self, mock_stdout_write):
        """ Test that the config file contains a list of files if multiple file arguments are passed. """
        self.assertRaises(SystemExit, parse_arguments)
        expected = "# Configuration file for Next-action. Edit the settings below as you like.\n"
        expected += "file:\n- ~/tasks.txt\n- project.txt\nnumber: 1\nreference: multiple\nstyle: default\n"
        self.assertEqual([call(expected)], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--write-config-file", "--all"])
    @patch.object(config, "open", mock_open(read_data=""))
    @patch.object(sys.stdout, "write")
    def test_show_all(self, mock_stdout_write):
        """ Test that the config file contains "all" true" instead of a number. """
        self.assertRaises(SystemExit, parse_arguments)
        expected = "# Configuration file for Next-action. Edit the settings below as you like.\n"
        expected += "all: true\nfile: ~/todo.txt\nreference: multiple\nstyle: default\n"
        self.assertEqual([call(expected)], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--write-config-file"])
    @patch.object(config, "open", mock_open(read_data="number: 3"))
    @patch.object(sys.stdout, "write")
    def test_read_config(self, mock_stdout_write):
        """ Test that the written config file contains the read config file. """
        self.assertRaises(SystemExit, parse_arguments)
        expected = "# Configuration file for Next-action. Edit the settings below as you like.\n"
        expected += "file: ~/todo.txt\nnumber: 3\nreference: multiple\nstyle: default\n"
        self.assertEqual([call(expected)], mock_stdout_write.call_args_list)


class FilenameTest(ConfigTestCase):
    """ Unit tests for the config file parameter. """

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="file: 0"))
    @patch.object(sys.stderr, "write")
    def test_invalid_filename(self, mock_stderr_write):
        """ Test a config file with an invalid file name. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE), call("next-action: error: ~/.next-action.cfg is invalid: "
                                       "file: must be of ['string', 'list'] type\n")],
            mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="file:\n- todo.txt\n- 0"))
    @patch.object(sys.stderr, "write")
    def test_valid_and_invalid(self, mock_stderr_write):
        """ Test a config file with an invalid file name. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE),
             call("next-action: error: ~/.next-action.cfg is invalid: file: 1: must be of string type\n")],
            mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="file: todo.txt"))
    def test_valid_file(self):
        """ Test that a valid filename changes the filenames. """
        self.assertEqual(["todo.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="file:\n- todo.txt\n- tada.txt"))
    def test_valid_files(self):
        """ Test that a list of valid filenames changes the filenames. """
        self.assertEqual(["todo.txt", "tada.txt"], parse_arguments()[1].file)

    @patch.object(sys, "argv", ["next-action", "--file", "tada.txt"])
    @patch.object(config, "open", mock_open(read_data="file: todo.txt"))
    def test_cli_takes_precedence(self):
        """ Test that a command line argument overrules the filename in the configuration file. """
        self.assertEqual(["tada.txt"], parse_arguments()[1].file)


class NumberTest(ConfigTestCase):
    """ Unit tests for the number and all parameters. """

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="number: not_a_number"))
    @patch.object(sys.stderr, "write")
    def test_invalid_number(self, mock_stderr_write):
        """ Test a config file with an invalid number. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE),
             call("next-action: error: ~/.next-action.cfg is invalid: number: must be of integer type\n")],
            mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="number: 0"))
    @patch.object(sys.stderr, "write")
    def test_zero(self, mock_stderr_write):
        """ Test a config file with an invalid number. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE),
             call("next-action: error: ~/.next-action.cfg is invalid: number: min value is 1\n")],
            mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="number: 3"))
    def test_valid_number(self):
        """ Test that a valid number changes the number argument. """
        self.assertEqual(3, parse_arguments()[1].number)

    @patch.object(sys, "argv", ["next-action", "--number", "3"])
    @patch.object(config, "open", mock_open(read_data="number: 2"))
    def test_cli_takes_precedence(self):
        """ Test that a command line argument overrules the number in the configuration file. """
        self.assertEqual(3, parse_arguments()[1].number)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="all: True"))
    def test_all_true(self):
        """ Test that all is true sets number to the max number. """
        self.assertEqual(sys.maxsize, parse_arguments()[1].number)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="all: False"))
    @patch.object(sys.stderr, "write")
    def test_all_false(self, mock_stderr_write):
        """ Test a config file with all is false. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE),
             call("next-action: error: ~/.next-action.cfg is invalid: all: unallowed value False\n")],
            mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="all: True\nnumber: 3"))
    @patch.object(sys.stderr, "write")
    def test_all_and_number(self, mock_stderr_write):
        """ Test that a config file with both --all and --number is invalid. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE),
             call("next-action: error: ~/.next-action.cfg is invalid: number: "
                  "'all' must not be present with 'number'\n")],
            mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--number", "3"])
    @patch.object(config, "open", mock_open(read_data="all: True"))
    def test_argument_nr_overrides(self):
        """ Test that --number on the command line overrides --all in the configuration file. """
        self.assertEqual(3, parse_arguments()[1].number)

    @patch.object(sys, "argv", ["next-action", "--all"])
    @patch.object(config, "open", mock_open(read_data="number: 3"))
    def test_argument_all_overrides(self):
        """ Test that --all on the command line overrides --number in the configuration file. """
        self.assertEqual(sys.maxsize, parse_arguments()[1].number)


class ConfigStyleTest(ConfigTestCase):
    """ Unit tests for the style parameter. """
    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="style: default"))
    def test_valid_style(self):
        """ Test that a valid style changes the style argument. """
        self.assertEqual("default", parse_arguments()[1].style)

    @patch.object(sys, "argv", ["next-action", "--style", "vim"])
    @patch.object(config, "open", mock_open(read_data="style: default"))
    def test_override_style(self):
        """ Test that a command line style overrides the style in the config file. """
        self.assertEqual("vim", parse_arguments()[1].style)

    @patch.object(sys, "argv", ["next-action", "--style"])
    @patch.object(config, "open", mock_open(read_data="style: default"))
    def test_cancel_style(self):
        """ Test that --style without style cancels the style in the config file. """
        self.assertEqual(None, parse_arguments()[1].style)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="style: invalid_style"))
    @patch.object(sys.stderr, "write")
    def test_invalid_style(self, mock_stderr_write):
        """ Test that an invalid style raises an error. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE),
             call("next-action: error: ~/.next-action.cfg is invalid: style: unallowed value invalid_style\n")],
            mock_stderr_write.call_args_list)


class PriorityTest(ConfigTestCase):
    """ Unit tests for the priority parameter. """

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="priority: Z"))
    def test_valid_priority(self):
        """ Test that a valid priority changes the priority argument. """
        self.assertEqual("Z", parse_arguments()[1].priority)

    @patch.object(sys, "argv", ["next-action", "--priority", "M"])
    @patch.object(config, "open", mock_open(read_data="priority: Z"))
    def test_override_priority(self):
        """ Test that a command line option overrides the priority in the config file. """
        self.assertEqual("M", parse_arguments()[1].priority)

    @patch.object(sys, "argv", ["next-action", "-pM"])
    @patch.object(config, "open", mock_open(read_data="priority: Z"))
    def test_override_short(self):
        """ Test that a short command line option overrides the priority in the config file. """
        self.assertEqual("M", parse_arguments()[1].priority)

    @patch.object(sys, "argv", ["next-action", "--priority"])
    @patch.object(config, "open", mock_open(read_data="priority: Z"))
    def test_cancel_priority(self):
        """ Test that a command line option overrides the priority in the config file. """
        self.assertEqual(None, parse_arguments()[1].priority)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="priority: ZZZ"))
    @patch.object(sys.stderr, "write")
    def test_invalid_priority(self, mock_stderr_write):
        """ Test that an invalid priority raises an error. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE),
             call("next-action: error: ~/.next-action.cfg is invalid: priority: unallowed value ZZZ\n")],
            mock_stderr_write.call_args_list)


class ReferenceTest(ConfigTestCase):
    """ Unit tests for the reference parameter. """

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="reference: always"))
    def test_valid_reference(self):
        """ Test that a valid reference value changes the reference argument. """
        self.assertEqual("always", parse_arguments()[1].reference)

    @patch.object(sys, "argv", ["next-action", "--reference", "never"])
    @patch.object(config, "open", mock_open(read_data="reference: always"))
    def test_override(self):
        """ Test that a command line argument overrides the configured value. """
        self.assertEqual("never", parse_arguments()[1].reference)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="reference: invalid"))
    @patch.object(sys.stderr, "write")
    def test_invalid_priority(self, mock_stderr_write):
        """ Test that an invalid value raises an error. """
        self.assertRaises(SystemExit, parse_arguments)
        self.assertEqual(
            [call(USAGE_MESSAGE),
             call("next-action: error: ~/.next-action.cfg is invalid: reference: unallowed value invalid\n")],
            mock_stderr_write.call_args_list)


class FiltersTest(ConfigTestCase):
    """ Unit tests for the filters parameter. """

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters: +ImportantProject"))
    def test_project(self):
        """ Test that a valid project changes the filters argument. """
        self.assertEqual({"ImportantProject"}, parse_arguments()[1].projects)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters: +ImportantProject +SideProject"))
    def test_projects(self):
        """ Test that multiple valid projects change the filters argument. """
        self.assertEqual({"ImportantProject", "SideProject"}, parse_arguments()[1].projects)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters: '@work'"))
    def test_context(self):
        """ Test that a valid context changes the filters argument. """
        self.assertEqual({"work"}, parse_arguments()[1].contexts)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters: '@work @desk'"))
    def test_contexts(self):
        """ Test that valid contexts change the filters argument. """
        self.assertEqual({"work", "desk"}, parse_arguments()[1].contexts)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters:\n- '@work'\n- -@waiting\n"))
    def test_context_list(self):
        """ Test that a valid context changes the filters argument. """
        self.assertEqual({"work"}, parse_arguments()[1].contexts)
        self.assertEqual({"waiting"}, parse_arguments()[1].excluded_contexts)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters: -+SideProject"))
    def test_excluded_project(self):
        """ Test that an excluded valid project changes the filters argument. """
        self.assertEqual({"SideProject"}, parse_arguments()[1].excluded_projects)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters: -@home"))
    def test_excluded_context(self):
        """ Test that an excluded valid context changes the filters argument. """
        self.assertEqual({"home"}, parse_arguments()[1].excluded_contexts)

    @patch.object(sys, "argv", ["next-action", "+ImportantProject"])
    @patch.object(config, "open", mock_open(read_data="filters: +ImportantProject"))
    def test_same_project(self):
        """ Test that the configuration is ignored when the command line specifies the same project. """
        self.assertEqual({"ImportantProject"}, parse_arguments()[1].projects)

    @patch.object(sys, "argv", ["next-action", "+ImportantProject"])
    @patch.object(config, "open", mock_open(read_data="filters: -+ImportantProject"))
    def test_inverse_project(self):
        """ Test that the configuration is ignored when the command line specifies the same project. """
        self.assertEqual(set(), parse_arguments()[1].excluded_projects)
        self.assertEqual({"ImportantProject"}, parse_arguments()[1].projects)

    @patch.object(sys, "argv", ["next-action", "+ImportantProject"])
    @patch.object(config, "open", mock_open(read_data="filters: -+ImportantProject @work"))
    def test_ignore_project_not_context(self):
        """ Test that the configuration is ignored when the command line specifies the same project. """
        self.assertEqual(set(), parse_arguments()[1].excluded_projects)
        self.assertEqual({"ImportantProject"}, parse_arguments()[1].projects)
        self.assertEqual({"work"}, parse_arguments()[1].contexts)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters:\n- invalid\n"))
    @patch.object(sys.stderr, "write")
    def test_invalid_filter_list(self, mock_stderr_write):
        """ Test that an invalid filter raises an error. """
        self.assertRaises(SystemExit, parse_arguments)
        regex = r"^\-?[@|\+]\S+"
        self.assertEqual(
            [call(USAGE_MESSAGE), call("next-action: error: ~/.next-action.cfg is invalid: filters: 0: "
                                       "value does not match regex '{0}'\n".format(regex))],
            mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch.object(config, "open", mock_open(read_data="filters: invalid\n"))
    @patch.object(sys.stderr, "write")
    def test_invalid_filter_string(self, mock_stderr_write):
        """ Test that an invalid filter raises an error. """
        self.assertRaises(SystemExit, parse_arguments)
        regex = r"^\-?[@|\+]\S+(\s+\-?[@|\+]\S+)*"
        self.assertEqual(
            [call(USAGE_MESSAGE), call("next-action: error: ~/.next-action.cfg is invalid: filters: "
                                       "value does not match regex '{0}'\n".format(regex))],
            mock_stderr_write.call_args_list)



""" Unit tests for the reference module in the output package. """

import unittest

from next_action import todotxt
from next_action.output import reference


class ReferenceTest(unittest.TestCase):
    """ Unit tests for the reference method. """
    def test_reference(self):
        """ Test that the source filename is added. """
        self.assertEqual("Todo [todo.txt]", reference(todotxt.Task("Todo", "todo.txt")))



""" Colorize the output. """

from pygments import lexers, formatters, highlight


def colorize(todotxt: str, style: str) -> str:
    """ Colorize the todotxt text according to the style. """
    if not style:
        return todotxt
    lexer = lexers.get_lexer_by_name("todotxt")
    formatter = formatters.get_formatter_by_name("terminal256", style=style)
    result = highlight(todotxt, lexer, formatter)
    if not todotxt.endswith("\n"):
        result = result[:-1]  # Remove the line break that Pygments added
    return result



""" Unit tests for the todo.txt Task class. """

import datetime
import string
import unittest

from hypothesis import given, strategies

from next_action import todotxt


class TodoTest(unittest.TestCase):
    """ Unit tests for the Task class that represents one task item. """

    def test_task_text(self):
        """ Test that the task text can be retrieved. """
        self.assertEqual("Todo", todotxt.Task("Todo").text)

    def test_task_repr(self):
        """ Test the task repr(). """
        self.assertEqual("Task<Todo>", repr(todotxt.Task("Todo")))


class TodoContextTest(unittest.TestCase):
    """ Unit tests for Task contexts. """

    def test_no_context(self):
        """ Test that a task without contexts has no contexts. """
        self.assertEqual(set(), todotxt.Task("Todo").contexts())

    def test_one_context(self):
        """ Test that the task context can be retrieved. """
        self.assertEqual({"home"}, todotxt.Task("Todo @home").contexts())

    def test_two_contexts(self):
        """ Test that multiple task contexts can be retrieved. """
        self.assertEqual({"home", "work"}, todotxt.Task("Todo @home @work").contexts())

    def test_no_space_before_at_sign(self):
        """ Test that a context with a space before the @-sign is not considered a context. """
        self.assertEqual(set(), todotxt.Task("Todo@home").contexts())


class TaskProjectTest(unittest.TestCase):
    """ Unit tests for task projects. """

    def test_no_projects(self):
        """ Test a task without projects. """
        self.assertEqual(set(), todotxt.Task("Todo").projects())

    def test_one_project(self):
        """ Test that the task project can be retrieved. """
        self.assertEqual({"projectx"}, todotxt.Task("Todo +projectx").projects())

    def test_two_projects(self):
        """ Test that multiple task projects can be retrieved. """
        self.assertEqual({"projectx", "projecty"}, todotxt.Task("Todo +projectx +projecty").projects())

    def test_no_space_before_at_sign(self):
        """ Test that a project with a space before the +-sign is not considered a project. """
        self.assertEqual(set(), todotxt.Task("Todo+project").projects())


class TaskPriorityTest(unittest.TestCase):
    """ Unit test for task priority. """

    def test_no_priority(self):
        """ Test a task without priority. """
        self.assertEqual(None, todotxt.Task("Todo").priority())

    def test_priorities(self):
        """ Test a task with a priority.  """
        for priority in string.ascii_uppercase:
            self.assertEqual(priority, todotxt.Task("({0}) Todo".format(priority)).priority())

    def test_faulty_priorities(self):
        """ Test that priorities must be one upper case letter. """
        for todo_txt in ("(a) Todo", " (a) Todo", "(A1) Todo", "(A Todo", "Todo (A)", "(A)Todo"):
            self.assertEqual(None, todotxt.Task(todo_txt).priority())

    def test_priority_at_least(self):
        """ Test that the task has a minimum priority. """
        self.assertTrue(todotxt.Task("(A) Task").priority_at_least("A"))
        self.assertTrue(todotxt.Task("(A) Task").priority_at_least("B"))
        self.assertFalse(todotxt.Task("(B) Task").priority_at_least("A"))
        self.assertFalse(todotxt.Task("Task").priority_at_least("Z"))
        self.assertTrue(todotxt.Task("(Z) Task").priority_at_least(None))
        self.assertTrue(todotxt.Task("Task").priority_at_least(None))

    def test_blocking(self):
        """ Test that the priority of a task without its own priority equals the priority of the task it is
            blocking. """
        tasks = todotxt.Tasks()
        after = todotxt.Task("(A) After id:1", tasks=tasks)
        before = todotxt.Task("Before before:1", tasks=tasks)
        tasks.extend([before, after])
        self.assertEqual("A", before.priority())

    def test_blocking_multiple(self):
        """ Test that the priority of a task without its own priority equals the highest priority of the tasks it is
            blocking. """
        tasks = todotxt.Tasks()
        after1 = todotxt.Task("(A) After id:1", tasks=tasks)
        after2 = todotxt.Task("(B) After after:before", tasks=tasks)
        before = todotxt.Task("Before before:1 id:before", tasks=tasks)
        tasks.extend([before, after1, after2])
        self.assertEqual("A", before.priority())

    def test_blocking_completed(self):
        """ Test that the priority of a completed blocked task is ignored. """
        tasks = todotxt.Tasks()
        after = todotxt.Task("(B) After id:1", tasks=tasks)
        after_completed = todotxt.Task("x (A) After id:2", tasks=tasks)
        before = todotxt.Task("Before before:1 before:2", tasks=tasks)
        tasks.extend([before, after, after_completed])
        self.assertEqual("B", before.priority())


class CreationDateTest(unittest.TestCase):
    """ Unit tests for task creation dates. Next-action interprets creation dates in the future as threshold, or
        start date. """

    def test_no_creation_date(self):
        """ Test that tasks have no creation date by default. """
        self.assertEqual(None, todotxt.Task("Todo").creation_date())

    def test_creation_date(self):
        """ Test a valid creation date. """
        self.assertEqual(datetime.date(2018, 1, 2), todotxt.Task("2018-01-02 Todo").creation_date())

    def test_creation_date_after_priority(self):  # pylint:disable=invalid-name
        """ Test a valid creation date after the priority. """
        self.assertEqual(datetime.date(2018, 12, 3), todotxt.Task("(B) 2018-12-03 Todo").creation_date())

    def test_invalid_creation_date(self):
        """ Test an invalid creation date. """
        self.assertEqual(None, todotxt.Task("2018-14-02 Todo").creation_date())

    def test_no_space_after(self):
        """ Test a creation date without a word boundary. """
        self.assertEqual(None, todotxt.Task("2018-10-10Todo").creation_date())

    def test_single_digits(self):
        """ Test a creation date with single digits for day and/or month. """
        self.assertEqual(datetime.date(2018, 12, 3), todotxt.Task("(B) 2018-12-3 Todo").creation_date())
        self.assertEqual(datetime.date(2018, 1, 13), todotxt.Task("(B) 2018-1-13 Todo").creation_date())
        self.assertEqual(datetime.date(2018, 1, 1), todotxt.Task("(B) 2018-1-1 Todo").creation_date())

    def test_is_future_task(self):
        """ Test that a task with a creation date in the future is a future task. """
        self.assertTrue(todotxt.Task("9999-01-01 Prepare for five-digit years").is_future())
        self.assertFalse(todotxt.Task("{0} Todo".format(datetime.date.today().isoformat())).is_future())


class ThresholdDateTest(unittest.TestCase):
    """ Unit tests for the threshold date of a task. The core todo.txt standard only defines creation date, but
        threshold (t:<date>) seems to be a widely used convention. """

    def test_no_threshold(self):
        """ Test that tasks have no threshold by default. """
        task = todotxt.Task("Todo")
        self.assertEqual(None, task.threshold_date())
        self.assertFalse(task.is_future())

    def test_past_threshold(self):
        """ Test a past threshold date. """
        task = todotxt.Task("Todo t:2018-01-02")
        self.assertEqual(datetime.date(2018, 1, 2), task.threshold_date())
        self.assertFalse(task.is_future())

    def test_future_threshold(self):
        """ Test a future threshold date. """
        task = todotxt.Task("Todo t:9999-01-01")
        self.assertEqual(datetime.date(9999, 1, 1), task.threshold_date())
        self.assertTrue(task.is_future())

    def test_threshold_today(self):
        """ Test a task with threshold today. """
        task = todotxt.Task("Todo t:{0}".format(datetime.date.today().isoformat()))
        self.assertEqual(datetime.date.today(), task.threshold_date())
        self.assertFalse(task.is_future())


class DueDateTest(unittest.TestCase):
    """ Unit tests for the due date of tasks. """

    def test_no_due_date(self):
        """ Test that tasks have no due date by default. """
        task = todotxt.Task("Todo")
        self.assertEqual(None, task.due_date())
        self.assertFalse(task.is_overdue())

    def test_past_due_date(self):
        """ Test a past due date. """
        task = todotxt.Task("Todo due:2018-01-02")
        self.assertEqual(datetime.date(2018, 1, 2), task.due_date())
        self.assertTrue(task.is_overdue())

    def test_future_due_date(self):
        """ Test a future due date. """
        task = todotxt.Task("Todo due:9999-01-01")
        self.assertEqual(datetime.date(9999, 1, 1), task.due_date())
        self.assertFalse(task.is_overdue())

    def test_due_today(self):
        """ Test a task due today. """
        task = todotxt.Task("Todo due:{0}".format(datetime.date.today().isoformat()))
        self.assertEqual(datetime.date.today(), task.due_date())
        self.assertFalse(task.is_overdue())

    def test_invalid_date(self):
        """ Test an invalid due date. """
        task = todotxt.Task("Todo due:2018-01-32")
        self.assertEqual(None, task.due_date())
        self.assertFalse(task.is_overdue())

    def test_no_space_after(self):
        """ Test a due date without a word boundary following it. """
        task = todotxt.Task("Todo due:2018-01-023")
        self.assertEqual(None, task.due_date())
        self.assertFalse(task.is_overdue())

    @given(strategies.sampled_from(["2018-01-1", "2018-1-01", "2018-1-1"]))
    def test_single_digits(self, due_date):
        """ Test a due date with single digits for day and/or month. """
        self.assertEqual(datetime.date(2018, 1, 1), todotxt.Task("(B) due:{0} Todo".format(due_date)).due_date())

    def test_is_due(self):
        """ Test the is_due method. """
        self.assertTrue(todotxt.Task("due:2018-01-01").is_due(datetime.date.today()))
        self.assertTrue(todotxt.Task("due:2018-01-01").is_due(datetime.date(2018, 1, 1)))
        self.assertTrue(todotxt.Task("9999-01-01 due:2018-01-01").is_due(datetime.date(2018, 1, 1)))
        self.assertTrue(todotxt.Task("x Completed due:2018-01-01").is_due(datetime.date(2018, 1, 1)))
        self.assertFalse(todotxt.Task("due:2018-01-01").is_due(datetime.date(2017, 12, 31)))
        self.assertFalse(todotxt.Task("Without due date").is_due(datetime.date(2017, 12, 31)))

    def test_blocking(self):
        """ Test that the due date of a task without its own due date equals the due date of the task it is
            blocking. """
        tasks = todotxt.Tasks()
        after = todotxt.Task("After id:1 due:2018-01-01", tasks=tasks)
        before = todotxt.Task("Before before:1", tasks=tasks)
        tasks.extend([before, after])
        self.assertEqual(datetime.date(2018, 1, 1), before.due_date())

    def test_blocking_multiple(self):
        """ Test that the due date of a task without its own due date equals the earliest due date of the tasks it is
            blocking. """
        tasks = todotxt.Tasks()
        after1 = todotxt.Task("After id:1 due:2018-10-01", tasks=tasks)
        after2 = todotxt.Task("After after:before due:2018-01-01", tasks=tasks)
        before = todotxt.Task("Before before:1 id:before", tasks=tasks)
        tasks.extend([before, after1, after2])
        self.assertEqual(datetime.date(2018, 1, 1), before.due_date())

    def test_blocking_completed(self):
        """ Test that the due date of a completed blocked task is ignored. """
        tasks = todotxt.Tasks()
        after = todotxt.Task("After id:1 due:2018-02-01", tasks=tasks)
        after_completed = todotxt.Task("x After id:2 due:2018-01-01", tasks=tasks)
        before = todotxt.Task("Before before:1 before:2", tasks=tasks)
        tasks.extend([before, after, after_completed])
        self.assertEqual(datetime.date(2018, 2, 1), before.due_date())


class TaskCompletionTest(unittest.TestCase):
    """ Unit tests for the completion status of tasks. """

    def test_not_completed(self):
        """ Test that a task that doesn't start with an x isn't completed. """
        self.assertFalse(todotxt.Task("Task x").is_completed())

    def test_completed(self):
        """ Test that a task that does start with an x is completed. """
        self.assertTrue(todotxt.Task("x Completed task").is_completed())

    def test_space_after_x(self):
        """ Test that a task that does start with an x is completed. """
        self.assertFalse(todotxt.Task("xNotCompleted").is_completed())

    def test_x_must_be_lowercase(self):
        """ Test that a task that starts with an X isn't completed. """
        self.assertFalse(todotxt.Task("X Not completed").is_completed())


class ActionableTest(unittest.TestCase):
    """ Unit tests for the actionable status of tasks. """

    def test_default(self):
        """ Test that a default task is actionable. """
        self.assertTrue(todotxt.Task("Todo").is_actionable())

    def test_past_creation_date(self):
        """ Test that a task with a past creation date is actionable. """
        self.assertTrue(todotxt.Task("2018-01-01 Todo").is_actionable())

    def test_future_creation_date(self):
        """ Test that a task with a future creation date is not actionable. """
        self.assertFalse(todotxt.Task("9999-01-01 Todo").is_actionable())

    def test_completed_task(self):
        """ Test that a completed task is not actionable. """
        self.assertFalse(todotxt.Task("x 2018-01-01 Todo").is_actionable())

    def test_two_dates(self):
        """ Test that a completed task with a past creation date is not actionable. """
        self.assertFalse(todotxt.Task("x 2018-01-01 2018-01-01 Todo").is_actionable())


class DependenciesTest(unittest.TestCase):
    """ Unit tests for dependency relations. """

    before_keys = strategies.sampled_from(["p", "before"])

    @given(strategies.sampled_from(["Todo", "Todo id:", "Todo id:id", "Todo p:", "Todo p:id",
                                    "Todo before:", "Todo before:id", "Todo after:", "Todo after:id"]))
    def test_task_without_dependencies(self, todo_text):
        """ Test that a task without dependencies is not blocked. """
        tasks = todotxt.Tasks()
        task = todotxt.Task(todo_text, tasks=tasks)
        another_task = todotxt.Task("Another task", tasks=tasks)
        tasks.extend([task, another_task])
        self.assertFalse(task.is_blocked())

    @given(before_keys)
    def test_one_before_another(self, before_key):
        """ Test that a task specified to be done before another task blocks the latter. """
        tasks = todotxt.Tasks()
        after = todotxt.Task("After id:1", tasks=tasks)
        before = todotxt.Task("Before {0}:1".format(before_key), tasks=tasks)
        tasks.extend([before, after])
        self.assertTrue(after.is_blocked())

    def test_one_after_another(self):
        """ Test that a task specified to be done after another task blocks the first. """
        tasks = todotxt.Tasks()
        after = todotxt.Task("After after:1", tasks=tasks)
        before = todotxt.Task("Before id:1", tasks=tasks)
        tasks.extend([before, after])
        self.assertTrue(after.is_blocked())

    @given(before_keys)
    def test_one_before_two(self, before_key):
        """ Test that a task that is specified to be done before two other tasks blocks both tasks. """
        tasks = todotxt.Tasks()
        before = todotxt.Task("Before {0}:1 {0}:2".format(before_key), tasks=tasks)
        after1 = todotxt.Task("After id:1", tasks=tasks)
        after2 = todotxt.Task("After id:1", tasks=tasks)
        tasks.extend([before, after1, after2])
        self.assertTrue(after1.is_blocked())
        self.assertTrue(after2.is_blocked())

    def test_one_after_two(self):
        """ Test that a task that is specified to be done after two other tasks is blocked by both. """
        tasks = todotxt.Tasks()
        before1 = todotxt.Task("Before 1 id:1", tasks=tasks)
        before2 = todotxt.Task("Before 2 id:2", tasks=tasks)
        after = todotxt.Task("After after:1 after:2", tasks=tasks)
        tasks.extend([before1, before2, after])
        self.assertTrue(after.is_blocked())

    @given(before_keys)
    def test_two_before_one(self, before_key):
        """ Test that a task that is specified to be done after two other tasks is blocked by both. """
        tasks = todotxt.Tasks()
        before1 = todotxt.Task("Before 1 {0}:1".format(before_key), tasks=tasks)
        before2 = todotxt.Task("Before 2 {0}:1".format(before_key), tasks=tasks)
        after = todotxt.Task("After id:1", tasks=tasks)
        tasks.extend([before1, before2, after])
        self.assertTrue(after.is_blocked())

    def test_two_after_one(self):
        """ Test that two tasks that are specified to be done after one other task are both blocked. """
        tasks = todotxt.Tasks()
        before = todotxt.Task("Before id:before", tasks=tasks)
        after1 = todotxt.Task("After 1 after:before", tasks=tasks)
        after2 = todotxt.Task("After 2 after:before", tasks=tasks)
        tasks.extend([before, after1, after2])
        self.assertTrue(after1.is_blocked())

    @given(before_keys)
    def test_completed_before_task(self, before_key):
        """ Test that a task is not blocked by a completed task. """
        tasks = todotxt.Tasks()
        after = todotxt.Task("After id:1", tasks=tasks)
        completed_before = todotxt.Task("x Before {0}:1".format(before_key), tasks=tasks)
        tasks.extend([after, completed_before])
        self.assertFalse(after.is_blocked())

    def test_after_completed_task(self):
        """ Test that a task is not blocked if it is specified to be done after a completed task. """
        tasks = todotxt.Tasks()
        after = todotxt.Task("After after:completed_before", tasks=tasks)
        completed_before = todotxt.Task("x Before id:completed_before", tasks=tasks)
        tasks.extend([after, completed_before])
        self.assertFalse(after.is_blocked())

    @given(before_keys)
    def test_self_before_self(self, before_key):
        """ Test that a task can be blocked by itself. This doesn't make sense, but we're not in the business
            of validating todo.txt files. """
        tasks = todotxt.Tasks()
        task = todotxt.Task("Todo id:1 {0}:1".format(before_key), tasks=tasks)
        tasks.append(task)
        self.assertTrue(task.is_blocked())

    def test_self_after_self(self):
        """ Test that a task can be blocked by itself. This doesn't make sense, but we're not in the business
            of validating todo.txt files. """
        tasks = todotxt.Tasks()
        task = todotxt.Task("Todo id:1 after:1", tasks=tasks)
        tasks.append(task)
        self.assertTrue(task.is_blocked())

    @given(before_keys)
    def test_self_before_self_indirect(self, before_key):
        """ Test that a task can be blocked by a second task that is blocked by the first task. This doesn't make sense,
            but we're not in the business of validating todo.txt files. """
        tasks = todotxt.Tasks()
        task1 = todotxt.Task("Task 1 id:1 {0}:2".format(before_key), tasks=tasks)
        task2 = todotxt.Task("Task 2 id:2 {0}:1".format(before_key), tasks=tasks)
        tasks.extend([task1, task2])
        self.assertTrue(task1.is_blocked())

    def test_self_after_self_indirect(self):
        """ Test that a task can be blocked by a second task that is blocked by the first task. This doesn't make sense,
            but we're not in the business of validating todo.txt files. """
        tasks = todotxt.Tasks()
        task1 = todotxt.Task("Task 1 id:1 after:2", tasks=tasks)
        task2 = todotxt.Task("Task 2 id:2 after:1", tasks=tasks)
        tasks.extend([task1, task2])
        self.assertTrue(task1.is_blocked())



""" Package for dealing with the todo.txt format. """

import fileinput
from typing import cast, List

from .task import Task
from .tasks import Tasks


def read_todotxt_files(filenames: List[str]) -> Tasks:
    """ Read tasks from the Todo.txt files. """
    tasks = Tasks()
    with cast(fileinput.FileInput, fileinput.input(filenames)) as todotxt_file:
        tasks.extend([Task(line.strip(), todotxt_file.filename(), tasks) for line in todotxt_file if line.strip()])
    return tasks



""" Reference the next actions with their source filename. """

from ..todotxt import Task


def reference(task: Task) -> str:
    """ Reference the next action with its source filename. """
    return "{0} [{1}]".format(task.text, task.filename)



""" Package for formatting output. """

import argparse

from ..todotxt import Task, Tasks

from .color import colorize
from .reference import reference


def render(tasks: Tasks, namespace: argparse.Namespace) -> str:
    """ Render the tasks using the options in the namespace. """
    reference_task = namespace.reference == "always" or namespace.reference == "multiple" and len(namespace.file) > 1
    render_task = reference if reference_task else lambda task: task.text
    return colorize("\n".join(render_task(task) for task in tasks), namespace.style or "")



""" Class that represents one task (i.e. one line) from a todo.txt file. """

import datetime
import re
import typing
from typing import List, Optional, Sequence, Set


class Task(object):
    """ A task from a line in a todo.txt file. """

    iso_date_reg_exp = r"(\d{4})-(\d{1,2})-(\d{1,2})"

    def __init__(self, todo_txt: str, filename: str = "", tasks: Sequence["Task"] = None) -> None:
        self.text = todo_txt
        self.filename = filename
        self.tasks = tasks if tasks is not None else []

    def __repr__(self) -> str:
        return "{0}<{1}>".format(self.__class__.__name__, self.text)

    def contexts(self) -> Set[str]:
        """ Return the contexts of the task. """
        return self.__prefixed_items("@")

    def projects(self) -> Set[str]:
        """ Return the projects of the task. """
        return self.__prefixed_items(r"\+")

    def priority(self) -> Optional[str]:
        """ Return the priority of the task """
        match = re.match(r"\(([A-Z])\) ", self.text)
        priorities: List[Optional[str]] = [match.group(1)] if match else []
        priorities.extend([blocked_task.priority() for blocked_task in self.__blocked_tasks()])
        return min(priorities, default=None, key=lambda priority: priority or "ZZZ")

    def priority_at_least(self, min_priority: Optional[str]) -> bool:
        """ Return whether the priority of task is at least the given priority. """
        if not min_priority:
            return True
        priority = self.priority()
        if not priority:
            return False
        return priority <= min_priority

    def creation_date(self) -> Optional[datetime.date]:
        """ Return the creation date of the task. """
        match = re.match(r"(?:\([A-Z]\) )?{0}\b".format(self.iso_date_reg_exp), self.text)
        return self.__create_date(match)

    def threshold_date(self) -> Optional[datetime.date]:
        """ Return the threshold date of the task. """
        return self.__find_keyed_date("t")

    def due_date(self) -> Optional[datetime.date]:
        """ Return the due date of the task. """
        due_dates = [self.__find_keyed_date("due")]
        due_dates.extend([blocked_task.due_date() for blocked_task in self.__blocked_tasks()])
        return min(due_dates, default=None, key=lambda due_date: due_date or datetime.date.max)

    def is_due(self, due_date: datetime.date) -> bool:
        """ Return whether the task is due on or before the given due date. """
        task_due_date = self.due_date()
        return task_due_date <= due_date if task_due_date else False

    def is_completed(self) -> bool:
        """ Return whether the task is completed or not. """
        return self.text.startswith("x ")

    def is_future(self) -> bool:
        """ Return whether the task is a future task, i.e. has a creation or threshold date in the future. """
        today = datetime.date.today()
        creation_date = self.creation_date()
        if creation_date:
            return creation_date > today
        threshold_date = self.threshold_date()
        if threshold_date:
            return threshold_date > today
        return False

    def is_actionable(self) -> bool:
        """ Return whether the task is actionable, i.e. whether it's not completed and doesn't have a future creation
            date. """
        return not self.is_completed() and not self.is_future()

    def is_overdue(self) -> bool:
        """ Return whether the task is overdue, i.e. whether it has a due date in the past. """
        due_date = self.due_date()
        return due_date < datetime.date.today() if due_date else False

    def is_blocked(self) -> bool:
        """ Return whether a task is blocked, i.e. whether it has uncompleted child tasks. """
        return any([task for task in self.tasks if not task.is_completed() and
                    (self.task_id() in task.parent_ids() or task.task_id() in self.child_ids())])

    def child_ids(self) -> Set[str]:
        """ Return the ids of the child tasks. """
        return {match.group(1) for match in re.finditer(r"\bafter:(\S+)\b", self.text)}

    def parent_ids(self) -> Set[str]:
        """ Return the ids of the parent tasks. """
        return {match.group(2) for match in re.finditer(r"\b(p|before):(\S+)\b", self.text)}

    def task_id(self) -> str:
        """ Return the id of the task. """
        match = re.search(r"\bid:(\S+)\b", self.text)
        return match.group(1) if match else ""

    def __blocked_tasks(self) -> Sequence["Task"]:
        """ Return the tasks this task is blocking. """
        return [task for task in self.tasks if not task.is_completed() and
                (task.task_id() in self.parent_ids() or self.task_id() in task.child_ids())]

    def __prefixed_items(self, prefix: str) -> Set[str]:
        """ Return the prefixed items in the task. """
        return {match.group(1) for match in re.finditer(" {0}([^ ]+)".format(prefix), self.text)}

    def __find_keyed_date(self, key: str) -> Optional[datetime.date]:
        """ Find a key:value pair with the supplied key where the value is a date. """
        match = re.search(r"\b{0}:{1}\b".format(key, self.iso_date_reg_exp), self.text)
        return self.__create_date(match)

    @staticmethod
    def __create_date(match: Optional[typing.Match[str]]) -> Optional[datetime.date]:
        """ Create a date from the match, if possible. """
        if match:
            try:
                return datetime.date(*(int(group) for group in match.groups()))
            except ValueError:
                pass
        return None



""" Parser for the command line arguments. """

import argparse
import datetime
import re
import shutil
import string
import sys
import textwrap
from typing import List, Set

import dateparser
from pygments.styles import get_all_styles

import next_action
from .config import read_config_file, write_config_file, validate_config_file


class NextActionArgumentParser(argparse.ArgumentParser):
    """ Command-line argument parser for Next-action. """

    def __init__(self) -> None:
        super().__init__(
            usage=textwrap.fill("next-action [-h] [--version] [-c [<config.cfg>] | -w] [-f <todo.txt> ...] "
                                "[-r <ref>] [-s [<style>]] [-a | -n <number>] [-d [<due date>] | -o] "
                                "[-p [<priority>]] [--] [<context|project> ...]",
                                width=shutil.get_terminal_size().columns - len("usage: ")),
            description="Show the next action in your todo.txt. The next action is selected from the tasks in the "
                        "todo.txt file based on task properties such as priority, due date, and creation date. Limit "
                        "the tasks from which the next action is selected by specifying contexts the tasks must have "
                        "and/or projects the tasks must belong to.",
            epilog="Use -- to separate options with optional arguments from contexts and projects, in order to handle "
                   "cases where a context or project is mistaken for an argument to an option.",
            formatter_class=CapitalisedHelpFormatter)
        self.__default_filenames = ["~/todo.txt"]
        self.add_optional_arguments()
        self.add_configuration_options()
        self.add_input_options()
        self.add_output_options()
        self.add_number_options()
        self.add_filter_arguments()

    def add_optional_arguments(self) -> None:
        """ Add the optional arguments to the parser. """
        self._optionals.title = self._optionals.title.capitalize() if self._optionals.title else None
        self.add_argument(
            "--version", action="version", version="%(prog)s {0}".format(next_action.__version__))

    def add_configuration_options(self) -> None:
        """ Add the configuration options to the parser. """
        config_group = self.add_argument_group("Configuration options")
        config_file = config_group.add_mutually_exclusive_group()
        config_file.add_argument(
            "-c", "--config-file", metavar="<config.cfg>", type=str, default="~/.next-action.cfg", nargs="?",
            help="filename of configuration file to read (default: %(default)s); omit filename to not read any "
                 "configuration file")
        config_file.add_argument(
            "-w", "--write-config-file", help="generate a sample configuration file and exit", action="store_true")

    def add_input_options(self) -> None:
        """ Add the input options to the parser. """
        input_group = self.add_argument_group("Input options")
        input_group.add_argument(
            "-f", "--file", action="append", metavar="<todo.txt>", default=self.__default_filenames[:], type=str,
            help="filename of todo.txt file to read; can be '-' to read from standard input; argument can be "
                 "repeated to read tasks from multiple todo.txt files (default: ~/todo.txt)")

    def add_output_options(self) -> None:
        """ Add the output/styling options to the parser. """
        output_group = self.add_argument_group("Output options")
        output_group.add_argument(
            "-r", "--reference", choices=["always", "never", "multiple"], default="multiple",
            help="reference next actions with the name of their todo.txt file (default: when reading multiple "
                 "todo.txt files)")
        styles = sorted(list(get_all_styles()))
        output_group.add_argument(
            "-s", "--style", metavar="<style>", choices=styles, default=None, nargs="?",
            help="colorize the output; available styles: {0} (default: %(default)s)".format(", ".join(styles)))

    def add_number_options(self) -> None:
        """ Add the number options to the parser. """
        number_group = self.add_argument_group("Show multiple next actions")
        number = number_group.add_mutually_exclusive_group()
        number.add_argument(
            "-a", "--all", default=1, action="store_const", dest="number", const=sys.maxsize,
            help="show all next actions")
        number.add_argument(
            "-n", "--number", metavar="<number>", type=int, default=1,
            help="number of next actions to show (default: %(default)s)")

    def add_filter_arguments(self) -> None:
        """ Add the filter arguments to the parser. """
        filters = self.add_argument_group("Limit the tasks from which the next actions are selected")
        date = filters.add_mutually_exclusive_group()
        date.add_argument(
            "-d", "--due", metavar="<due date>", type=date_type, nargs="?", const=datetime.date.max,
            help="show only next actions with a due date; if a date is given, show only next actions due on or "
                 "before that date")
        date.add_argument("-o", "--overdue", help="show only overdue next actions", action="store_true")
        filters.add_argument(
            "-p", "--priority", metavar="<priority>", choices=string.ascii_uppercase, nargs="?",
            help="minimum priority (A-Z) of next actions to show (default: %(default)s)")
        # Collect all context and project arguments in one list:
        filters.add_argument(
            "filters", metavar="<context|project>", help=argparse.SUPPRESS, nargs="*", type=filter_type)
        # Add two dummy arguments for the help info:
        filters.add_argument(
            "dummy", metavar="@<context> ...", nargs="*", default=argparse.SUPPRESS,
            help="contexts the next action must have")
        filters.add_argument(
            "dummy", metavar="+<project> ...", nargs="*", default=argparse.SUPPRESS,
            help="projects the next action must be part of; if repeated the next action must be part of at least one "
                 "of the projects")
        # These arguments won't be collected because they start with a -. They'll be parsed by parse_remaining_args
        # below
        filters.add_argument(
            "dummy", metavar="-@<context> ...", nargs="*", default=argparse.SUPPRESS,
            help="contexts the next action must not have")
        filters.add_argument(
            "dummy", metavar="-+<project> ...", nargs="*", default=argparse.SUPPRESS,
            help="projects the next action must not be part of")

    def parse_args(self, args=None, namespace=None) -> argparse.Namespace:
        """ Parse the command-line arguments. """
        namespace, remaining = self.parse_known_args(args, namespace)
        self.parse_remaining_args(remaining, namespace)
        if getattr(namespace, "config_file", self.get_default("config_file")) is not None:
            self.process_config_file(namespace)
        self.fix_filenames(namespace)
        if namespace.write_config_file:
            write_config_file(namespace)
            self.exit()
        return namespace

    def parse_remaining_args(self, remaining: List[str], namespace: argparse.Namespace) -> None:
        """ Parse the remaining command line arguments. """
        for value in remaining:
            if value.startswith("-@") or value.startswith("-+"):
                argument = value[len("-"):]
                if not argument[len("@"):]:
                    argument_type = "context" if argument.startswith("@") else "project"
                    self.error("argument <context|project>: {0} name missing".format(argument_type))
                elif argument in namespace.filters:
                    self.error("{0} is both included and excluded".format(argument))
                else:
                    namespace.filters.append(value)
            else:
                self.error("unrecognized arguments: {0}".format(value))
        namespace.contexts = subset(namespace.filters, "@")
        namespace.projects = subset(namespace.filters, "+")
        namespace.excluded_contexts = subset(namespace.filters, "-@")
        namespace.excluded_projects = subset(namespace.filters, "-+")

    def process_config_file(self, namespace: argparse.Namespace) -> None:
        """ Process the configuration file. """
        config_filename = namespace.config_file
        config = read_config_file(config_filename, self.get_default("config_file"), self.error)
        if not config:
            return
        validate_config_file(config, config_filename, self.error)
        self.insert_config(config, namespace)

    def insert_config(self, config, namespace: argparse.Namespace) -> None:
        """ Insert the configured parameters in the namespace, if no command line arguments are present. """
        if self.arguments_not_specified("-f", "--file"):
            filenames = config.get("file", [])
            if isinstance(filenames, str):
                filenames = [filenames]
            getattr(namespace, "file").extend(filenames)
        if self.arguments_not_specified("-n", "--number", "-a", "--all"):
            number = sys.maxsize if config.get("all", False) else config.get("number", self.get_default("number"))
            setattr(namespace, "number", number)
        if self.arguments_not_specified("-r", "--reference"):
            reference = config.get("reference", self.get_default("reference"))
            setattr(namespace, "reference", reference)
        if self.arguments_not_specified("-s", "--style"):
            style = config.get("style", self.get_default("style"))
            setattr(namespace, "style", style)
        if self.arguments_not_specified("-p", "--priority"):
            priority = config.get("priority", self.get_default("priority"))
            setattr(namespace, "priority", priority)
        self.insert_configured_filters(config, namespace)

    def insert_configured_filters(self, config, namespace: argparse.Namespace) -> None:
        """ Insert the configured filters in the namespace, if no matching command line filters are present. """
        filters = config.get("filters", [])
        if isinstance(filters, str):
            filters = re.split(r"\s", filters)
        for prefix, filter_key in (("@", "contexts"), ("+", "projects"),
                                   ("-@", "excluded_contexts"), ("-+", "excluded_projects")):
            for configured_filter in subset(filters, prefix):
                if self.filter_not_specified(prefix + configured_filter):
                    getattr(namespace, filter_key).add(configured_filter)

    @staticmethod
    def arguments_not_specified(*arguments: str) -> bool:
        """ Return whether any of the arguments was specified on the command line. """
        return not any([command_line_arg.startswith(argument) for argument in arguments
                        for command_line_arg in sys.argv])

    @staticmethod
    def filter_not_specified(filtered: str) -> bool:
        """ Return whether the context or project or its opposite were specified on the command line. """
        prefix = "-"
        opposite = filtered[len(prefix):] if filtered.startswith(prefix) else prefix + filtered
        return not (filtered in sys.argv or opposite in sys.argv)

    def fix_filenames(self, namespace: argparse.Namespace) -> None:
        """ Fix the filenames. """
        # Work around the issue that the "append" action doesn't overwrite defaults.
        # See https://bugs.python.org/issue16399.
        filenames = namespace.file[:]
        default_filenames = self.__default_filenames
        if default_filenames != filenames:
            for default_filename in default_filenames:
                filenames.remove(default_filename)
        # Remove duplicate filenames while maintaining order.
        namespace.file = list(dict.fromkeys(filenames))


class CapitalisedHelpFormatter(argparse.HelpFormatter):
    """ Capitalise the usage string. """
    def add_usage(self, usage, actions, groups, prefix=None):
        return super().add_usage(usage, actions, groups, prefix or "Usage: ")


def filter_type(value: str) -> str:
    """ Return the filter if it's valid, else raise an error. """
    if value.startswith("@") or value.startswith("+"):
        if value[len("@"):]:
            return value
        else:
            value_type = "context" if value.startswith("@") else "project"
            raise argparse.ArgumentTypeError("{0} name missing".format(value_type))
    raise argparse.ArgumentTypeError("unrecognized arguments: {0}".format(value))


def date_type(value: str) -> datetime.date:
    """ Return the date if it's valid, else raise an error. """
    date_time = dateparser.parse(value, languages=["en"],
                                 settings={"PREFER_DAY_OF_MONTH": "last", "PREFER_DATES_FROM": "future"})
    if date_time:
        return date_time.date()
    raise argparse.ArgumentTypeError("invalid date: {0}".format(value))


def subset(filters: List[str], prefix: str) -> Set[str]:
    """ Return a subset of the filters based on prefix. """
    return set([f.strip(prefix) for f in filters if f.startswith(prefix)])



""" Unit tests for the Tasks class. """

import unittest

from next_action.todotxt import Tasks, Task


class ContextsTest(unittest.TestCase):
    """ Unit tests for the contexts methods. """

    def test_no_tasks(self):
        """ Test that an empty tasks collection has no contexts. """
        self.assertEqual(set(), Tasks().contexts())

    def test_task_without_context(self):
        """ Test that a task collection with tasks without contexts has not contexts. """
        self.assertEqual(set(), Tasks([Task("Todo")]).contexts())

    def test_task_with_context(self):
        """ Test that a task collection with a task with a context has that context. """
        self.assertEqual(set(["work"]), Tasks([Task("Todo @work")]).contexts())

    def test_task_with_contexts(self):
        """ Test that a task collection with a task with contexts has those contexts. """
        self.assertEqual(set(["office", "work"]), Tasks([Task("Todo @work @office")]).contexts())

    def test_tasks_with_contexts(self):
        """ Test that a task collection with tasks with contexts has those contexts. """
        self.assertEqual(set(["office", "work", "home"]),
                         Tasks([Task("Todo @work @office"), Task("Todo @home @office")]).contexts())


class ProjectsTest(unittest.TestCase):
    """ Unit tests for the projects method. """

    def test_no_tasks(self):
        """ Test that an empty tasks collection has no projects. """
        self.assertEqual(set(), Tasks().projects())

    def test_task_without_project(self):
        """ Test that a task collection with tasks without projects has not projects. """
        self.assertEqual(set(), Tasks([Task("Todo")]).projects())

    def test_task_with_project(self):
        """ Test that a task collection with a task with a project has that project. """
        self.assertEqual(set(["BigProject"]), Tasks([Task("Todo +BigProject")]).projects())

    def test_task_with_projects(self):
        """ Test that a task collection with a task with projects has those projects. """
        self.assertEqual(set(["This", "That"]), Tasks([Task("Todo +This +That")]).projects())

    def test_tasks_with_projects(self):
        """ Test that a task collection with tasks with projects has those projects. """
        self.assertEqual(set(["Foo", "Bar", "Baz"]),
                         Tasks([Task("Todo +Foo +Bar"), Task("Todo +Foo +Baz")]).projects())



""" Unit test for the next action algorithm. """

import argparse
import datetime
import unittest
import sys

from next_action import todotxt, pick_action


class PickActionTestCase(unittest.TestCase):
    """ Base class for pick action unit tests. """

    def setUp(self):
        self.namespace = argparse.Namespace()
        self.namespace.contexts = set()
        self.namespace.projects = set()
        self.namespace.excluded_contexts = set()
        self.namespace.excluded_projects = set()
        self.namespace.overdue = False
        self.namespace.due = None
        self.namespace.priority = None
        self.namespace.number = sys.maxsize


class PickActionTest(PickActionTestCase):
    """ Unit tests for the pick action algorithm. """

    def test_no_tasks(self):
        """ Test that no tasks means no next action. """
        self.assertEqual([], pick_action.next_actions([], self.namespace))

    def test_one_task(self):
        """ If there is one task, that one is the next action. """
        task = todotxt.Task("Todo")
        self.assertEqual([task], pick_action.next_actions([task], self.namespace))

    def test_multiple_tasks(self):
        """ If there are multiple tasks, the first is the next action. """
        task1 = todotxt.Task("Todo 1")
        task2 = todotxt.Task("Todo 2")
        self.assertEqual([task1, task2], pick_action.next_actions([task1, task2], self.namespace))

    def test_higher_prio_goes_first(self):
        """ If there are multiple tasks with different priorities, the task with the
            highest priority is the next action. """
        task1 = todotxt.Task("Todo 1")
        task2 = todotxt.Task("(B) Todo 2")
        task3 = todotxt.Task("(A) Todo 3")
        self.assertEqual([task3, task2, task1], pick_action.next_actions([task1, task2, task3], self.namespace))

    def test_creation_dates(self):
        """ Test that a task with an older creation date takes precedence. """
        no_creation_date = todotxt.Task("Task 1")
        newer_task = todotxt.Task("2018-02-02 Task 2")
        older_task = todotxt.Task("2017-01-01 Task 3")
        self.assertEqual([older_task, newer_task, no_creation_date],
                         pick_action.next_actions([no_creation_date, newer_task, older_task], self.namespace))

    def test_priority_and_creation_date(self):
        """ Test that priority takes precedence over creation date. """
        priority = todotxt.Task("(C) Task 1")
        newer_task = todotxt.Task("2018-02-02 Task 2")
        older_task = todotxt.Task("2017-01-01 Task 3")
        self.assertEqual([priority, older_task, newer_task],
                         pick_action.next_actions([priority, newer_task, older_task], self.namespace))

    def test_due_dates(self):
        """ Test that a task with an earlier due date takes precedence. """
        no_due_date = todotxt.Task("Task 1")
        earlier_task = todotxt.Task("Task 2 due:2018-02-02")
        later_task = todotxt.Task("Task 3 due:2019-01-01")
        self.assertEqual([earlier_task, later_task, no_due_date],
                         pick_action.next_actions([no_due_date, later_task, earlier_task], self.namespace))

    def test_due_and_creation_dates(self):
        """ Test that a task with a due date takes precedence over creation date. """
        task1 = todotxt.Task("2018-1-1 Task 1")
        task2 = todotxt.Task("Task 2 due:2018-1-1")
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2], self.namespace))

    def test_project(self):
        """ Test that a task with a project takes precedence over a task without a project. """
        task1 = todotxt.Task("Task 1")
        task2 = todotxt.Task("Task 2 +Project")
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2], self.namespace))

    def test_projects(self):
        """ Test that a task with more projects takes precedence over a task with less projects. """
        task1 = todotxt.Task("Task 1 +Project")
        task2 = todotxt.Task("Task 2 +Project1 +Project2")
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2], self.namespace))


class FilterTasksTest(PickActionTestCase):
    """ Test that the tasks from which the next action is picked, can be filtered. """

    def test_context(self):
        """ Test that the next action can be limited to a specific context. """
        task1 = todotxt.Task("Todo 1 @work @computer")
        task2 = todotxt.Task("(B) Todo 2 @work")
        task3 = todotxt.Task("(A) Todo 3 @home")
        self.namespace.contexts = {"work"}
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2, task3], self.namespace))

    def test_contexts(self):
        """ Test that the next action can be limited to a set of contexts. """
        task1 = todotxt.Task("Todo 1 @work @computer")
        task2 = todotxt.Task("(B) Todo 2 @work @computer")
        task3 = todotxt.Task("(A) Todo 3 @home @computer")
        self.namespace.contexts = {"work", "computer"}
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2, task3], self.namespace))

    def test_excluded_context(self):
        """ Test that contexts can be excluded. """
        task = todotxt.Task("(A) Todo @computer")
        self.namespace.excluded_contexts = {"computer"}
        self.assertEqual([], pick_action.next_actions([task], self.namespace))

    def test_excluded_contexts(self):
        """ Test that contexts can be excluded. """
        task = todotxt.Task("(A) Todo @computer @work")
        task = todotxt.Task("(A) Todo @computer")
        self.namespace.excluded_contexts = {"computer"}
        self.assertEqual([], pick_action.next_actions([task], self.namespace))

    def test_not_excluded_context(self):
        """ Test that a task is not excluded if it doesn't belong to the excluded category. """
        task = todotxt.Task("(A) Todo @computer")
        self.namespace.excluded_contexts = {"phone"}
        self.assertEqual([task], pick_action.next_actions([task], self.namespace))

    def test_project(self):
        """ Test that the next action can be limited to a specific project. """
        task1 = todotxt.Task("Todo 1 +ProjectX")
        task2 = todotxt.Task("(B) Todo 2 +ProjectX")
        task3 = todotxt.Task("(A) Todo 3 +ProjectY")
        self.namespace.projects = {"ProjectX"}
        self.assertEqual([task2, task1], pick_action.next_actions([task1, task2, task3], self.namespace))

    def test_excluded_project(self):
        """ Test that projects can be excluded. """
        task = todotxt.Task("(A) Todo +DogHouse")
        self.namespace.excluded_projects = {"DogHouse"}
        self.assertEqual([], pick_action.next_actions([task], self.namespace))

    def test_excluded_projects(self):
        """ Test that projects can be excluded. """
        task = todotxt.Task("(A) Todo +DogHouse +PaintHouse")
        self.namespace.excluded_projects = {"DogHouse"}
        self.assertEqual([], pick_action.next_actions([task], self.namespace))

    def test_not_excluded_project(self):
        """ Test that a task is not excluded if it doesn't belong to the excluded project. """
        task = todotxt.Task("(A) Todo +DogHouse")
        self.namespace.excluded_projects = {"PaintHouse"}
        self.assertEqual([task], pick_action.next_actions([task], self.namespace))

    def test_project_and_context(self):
        """ Test that the next action can be limited to a specific project and context. """
        task1 = todotxt.Task("Todo 1 +ProjectX @office")
        task2 = todotxt.Task("(B) Todo 2 +ProjectX")
        task3 = todotxt.Task("(A) Todo 3 +ProjectY")
        self.namespace.contexts = {"office"}
        self.namespace.projects = {"ProjectX"}
        self.assertEqual([task1], pick_action.next_actions([task1, task2, task3], self.namespace))


class IgnoredTasksTest(PickActionTestCase):
    """ Test that certain tasks are ignored when picking the next action. """

    def test_ignore_completed_task(self):
        """ If there's one completed and one uncompleted task, the uncompleted one is the next action. """
        completed_task = todotxt.Task("x Completed")
        uncompleted_task = todotxt.Task("Todo")
        self.assertEqual(
            [uncompleted_task], pick_action.next_actions([completed_task, uncompleted_task], self.namespace))

    def test_ignore_future_task(self):
        """ Ignore tasks with a start date in the future. """
        future_task = todotxt.Task("(A) 9999-01-01 Start preparing for five-digit years")
        regular_task = todotxt.Task("(B) Look busy")
        self.assertEqual([regular_task], pick_action.next_actions([future_task, regular_task], self.namespace))

    def test_ignore_these_tasks(self):
        """ If all tasks are completed or future tasks, there's no next action. """
        completed_task1 = todotxt.Task("x Completed")
        completed_task2 = todotxt.Task("x Completed too")
        future_task = todotxt.Task("(A) 9999-01-01 Start preparing for five-digit years")
        self.assertEqual([], pick_action.next_actions([completed_task1, completed_task2, future_task], self.namespace))


class OverdueTasks(PickActionTestCase):
    """ Unit tests for the overdue filter. """
    def test_overdue_tasks(self):
        """ Test that tasks that not overdue are filtered. """
        no_duedate = todotxt.Task("Task")
        future_duedate = todotxt.Task("Task due:9999-01-01")
        overdue = todotxt.Task("Task due:2000-01-01")
        self.namespace.overdue = True
        self.assertEqual([overdue], pick_action.next_actions([no_duedate, future_duedate, overdue], self.namespace))


class DueTasks(PickActionTestCase):
    """ Unit tests for the is_due filter. """
    def test_due_tasks(self):
        """ Test that tasks that are not due are filtered. """
        no_duedate = todotxt.Task("Task")
        future_duedate = todotxt.Task("Task due:9999-01-01")
        overdue = todotxt.Task("Task due:2000-01-01")
        self.namespace.due = datetime.date(2000, 1, 1)
        self.assertEqual([overdue], pick_action.next_actions([no_duedate, future_duedate, overdue], self.namespace))

    def test_any_due_tasks(self):
        """ Test that tasks that are not due are filtered. """
        no_duedate = todotxt.Task("Task")
        future_duedate = todotxt.Task("Task due:9999-01-01")
        overdue = todotxt.Task("Task due:2000-01-01")
        self.namespace.due = datetime.date.max
        self.assertEqual([overdue, future_duedate],
                         pick_action.next_actions([no_duedate, future_duedate, overdue], self.namespace))


class MinimimPriorityTest(PickActionTest):
    """ Unit test for the mininum priority filter. """

    def test_priority(self):
        """ Test that tasks without priority are filtered. """
        no_priority = todotxt.Task("Task")
        high_priority = todotxt.Task("(A) Task")
        low_priority = todotxt.Task("(Z) Task")
        self.namespace.priority = "K"
        self.assertEqual(
            [high_priority], pick_action.next_actions([no_priority, low_priority, high_priority], self.namespace))


class BlockedTasksTest(PickActionTest):
    """ Unit tests for blocked tasks. """

    def test_blocked(self):
        """ Test that a blocked task isn't the next action. """
        tasks = todotxt.Tasks()
        parent = todotxt.Task("(A) Parent id:1", tasks=tasks)
        child = todotxt.Task("Child p:1", tasks=tasks)
        tasks.extend([child, parent])
        self.assertEqual([child], pick_action.next_actions(tasks, self.namespace))

    def test_blocked_chain(self):
        """ Test that a blocked task isn't the next action. """
        tasks = todotxt.Tasks()
        grandparent = todotxt.Task("(A) Grandparent id:1", tasks=tasks)
        parent = todotxt.Task("(B) Parent p:1 id:2", tasks=tasks)
        child = todotxt.Task("Child p:2", tasks=tasks)
        tasks.extend([child, parent, grandparent])
        self.assertEqual([child], pick_action.next_actions(tasks, self.namespace))

    def test_multiple_childs(self):
        """ Test that a blocked task isn't the next action. """
        tasks = todotxt.Tasks()
        parent = todotxt.Task("(A) Parent id:1", tasks=tasks)
        child1 = todotxt.Task("x Child 1 p:1", tasks=tasks)
        child2 = todotxt.Task("Child 2 p:1", tasks=tasks)
        child3 = todotxt.Task("Child 3 p:1", tasks=tasks)
        tasks.extend([parent, child1, child2, child3])
        self.assertEqual([child2, child3], pick_action.next_actions(tasks, self.namespace))

    def test_multiple_parents(self):
        """ Test that a blocked task isn't the next action. """
        tasks = todotxt.Tasks()
        parent1 = todotxt.Task("(A) Parent id:1", tasks=tasks)
        parent2 = todotxt.Task("(B) Parent id:2", tasks=tasks)
        child = todotxt.Task("Child 1 p:1 p:2", tasks=tasks)
        tasks.extend([parent1, parent2, child])
        self.assertEqual([child], pick_action.next_actions(tasks, self.namespace))

    def test_due_date_of_blocked_task(self):
        """ Test that a task that blocks a task with an earlier due date takes precendence. """
        tasks = todotxt.Tasks()
        due_first_but_blocked = todotxt.Task("Task id:1 due:2018-01-01", tasks=tasks)
        blocking_task = todotxt.Task("Blocking before:1", tasks=tasks)
        due_second = todotxt.Task("Task due:2018-02-01", tasks=tasks)
        tasks.extend([due_first_but_blocked, blocking_task, due_second])
        self.assertEqual([blocking_task, due_second], pick_action.next_actions(tasks, self.namespace))

    def test_priority_of_blocked_task(self):
        """ Test that a task that blocks a task with a higher priority takes precendence. """
        tasks = todotxt.Tasks()
        high_prio_but_blocked = todotxt.Task("(A) Task id:1", tasks=tasks)
        blocking_task = todotxt.Task("Blocking before:1", tasks=tasks)
        second_prio = todotxt.Task("(B) Task", tasks=tasks)
        tasks.extend([high_prio_but_blocked, blocking_task, second_prio])
        self.assertEqual([blocking_task, second_prio], pick_action.next_actions(tasks, self.namespace))



""" Setuptools setup script for Next-action. """

from setuptools import setup, find_packages

import next_action


setup(
    name=next_action.__title__,
    version=next_action.__version__,
    description="Command-line application to show the next action to work on from a todo.txt file",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Frank Niessink",
    author_email="frank@niessink.com",
    url="https://github.com/fniessink/next-action",
    license="Apache License, Version 2.0",
    python_requires=">=3.6",
    install_requires=["dateparser", "Cerberus", "PyYAML", "Pygments"],
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "next-action=next_action.cli:next_action",
        ],
    },
    test_suite="tests",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Topic :: Office/Business :: Scheduling"],
    keywords=["todo.txt", "task management", "todolist"])



""" Unit tests for the command-line interface entry point. """

import os
import sys
import unittest
from unittest.mock import patch, mock_open, call

from next_action.cli import next_action
from next_action.arguments import config
from next_action import __version__

from .arguments.test_parser import USAGE_MESSAGE


@patch.object(config, "open", mock_open(read_data=""))
class CLITest(unittest.TestCase):
    """ Unit tests for the command-line interface. """

    @patch.object(sys, "argv", ["next-action"])
    @patch("fileinput.open", mock_open(read_data=""))
    @patch.object(sys.stdout, "write")
    def test_empty_task_file(self, mock_stdout_write):
        """ Test the response when the task file is empty. """
        next_action()
        self.assertEqual([call("Nothing to do!"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch("fileinput.open", mock_open(read_data="Todo\n"))
    @patch.object(sys.stdout, "write")
    def test_one_task(self, mock_stdout_write):
        """ Test the response when the task file has one task. """
        next_action()
        self.assertEqual([call("Todo"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "@work"])
    @patch("fileinput.open", mock_open(read_data="Todo @home\nTodo @work\n"))
    @patch.object(sys.stdout, "write")
    def test_context(self, mock_stdout_write):
        """ Test the response when the user passes a context. """
        next_action()
        self.assertEqual([call("Todo @work"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "+DogHouse"])
    @patch("fileinput.open", mock_open(read_data="Walk the dog @home\nBuy wood +DogHouse\n"))
    @patch.object(sys.stdout, "write")
    def test_project(self, mock_stdout_write):
        """ Test the response when the user passes a project. """
        next_action()
        self.assertEqual([call("Buy wood +DogHouse"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action"])
    @patch("fileinput.open")
    @patch.object(sys.stderr, "write")
    def test_missing_file(self, mock_stderr_write, mock_file_open):
        """ Test the response when the task file can't be found. """
        mock_file_open.side_effect = OSError("some problem")
        self.assertRaises(SystemExit, next_action)
        self.assertEqual([call(USAGE_MESSAGE), call("next-action: error: can't open file: some problem\n")],
                         mock_stderr_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--help"])
    @patch.object(sys.stdout, "write")
    def test_help(self, mock_stdout_write):
        """ Test the help message. """
        os.environ['COLUMNS'] = "120"  # Fake that the terminal is wide enough.
        self.assertRaises(SystemExit, next_action)
        self.assertEqual(call("""\
Usage: next-action [-h] [--version] [-c [<config.cfg>] | -w] [-f <todo.txt> ...] [-r <ref>] [-s [<style>]] [-a | -n
<number>] [-d [<due date>] | -o] [-p [<priority>]] [--] [<context|project> ...]

Show the next action in your todo.txt. The next action is selected from the tasks in the todo.txt file based on task
properties such as priority, due date, and creation date. Limit the tasks from which the next action is selected by
specifying contexts the tasks must have and/or projects the tasks must belong to.

Optional arguments:
  -h, --help            show this help message and exit
  --version             show program's version number and exit

Configuration options:
  -c [<config.cfg>], --config-file [<config.cfg>]
                        filename of configuration file to read (default: ~/.next-action.cfg); omit filename to not
                        read any configuration file
  -w, --write-config-file
                        generate a sample configuration file and exit

Input options:
  -f <todo.txt>, --file <todo.txt>
                        filename of todo.txt file to read; can be '-' to read from standard input; argument can be
                        repeated to read tasks from multiple todo.txt files (default: ~/todo.txt)

Output options:
  -r {always,never,multiple}, --reference {always,never,multiple}
                        reference next actions with the name of their todo.txt file (default: when reading multiple
                        todo.txt files)
  -s [<style>], --style [<style>]
                        colorize the output; available styles: abap, algol, algol_nu, arduino, autumn, borland, bw,
                        colorful, default, emacs, friendly, fruity, igor, lovelace, manni, monokai, murphy, native,
                        paraiso-dark, paraiso-light, pastie, perldoc, rainbow_dash, rrt, tango, trac, vim, vs, xcode
                        (default: None)

Show multiple next actions:
  -a, --all             show all next actions
  -n <number>, --number <number>
                        number of next actions to show (default: 1)

Limit the tasks from which the next actions are selected:
  -d [<due date>], --due [<due date>]
                        show only next actions with a due date; if a date is given, show only next actions due on or
                        before that date
  -o, --overdue         show only overdue next actions
  -p [<priority>], --priority [<priority>]
                        minimum priority (A-Z) of next actions to show (default: None)
  @<context> ...        contexts the next action must have
  +<project> ...        projects the next action must be part of; if repeated the next action must be part of at least
                        one of the projects
  -@<context> ...       contexts the next action must not have
  -+<project> ...       projects the next action must not be part of

Use -- to separate options with optional arguments from contexts and projects, in order to handle cases where a
context or project is mistaken for an argument to an option.
"""),
                         mock_stdout_write.call_args_list[0])

    @patch.object(sys, "argv", ["next-action", "--version"])
    @patch.object(sys.stdout, "write")
    def test_version(self, mock_stdout_write):
        """ Test that --version shows the version number. """
        self.assertRaises(SystemExit, next_action)
        self.assertEqual([call("next-action {0}\n".format(__version__))], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--number", "2"])
    @patch("fileinput.open", mock_open(read_data="Walk the dog @home\n(A) Buy wood +DogHouse\n(B) Call mom\n"))
    @patch.object(sys.stdout, "write")
    def test_number(self, mock_stdout_write):
        """ Test that the number of next actions can be specified. """
        next_action()
        self.assertEqual([call("(A) Buy wood +DogHouse\n(B) Call mom"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--number", "3"])
    @patch("fileinput.open", mock_open(read_data="\nWalk the dog @home\n   \n(B) Call mom\n"))
    @patch.object(sys.stdout, "write")
    def test_ignore_empty_lines(self, mock_stdout_write):
        """ Test that empty lines in the todo.txt file are ignored. """
        next_action()
        self.assertEqual([call("(B) Call mom\nWalk the dog @home"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--all"])
    @patch("fileinput.open", mock_open(read_data="\nWalk the dog @home\nBuy beer\n(B) Call mom\n"))
    @patch.object(sys.stdout, "write")
    def test_show_all_actions(self, mock_stdout_write):
        """ Test that all actions in the todo.txt file are shown. """
        next_action()
        self.assertEqual([call("(B) Call mom\nWalk the dog @home\nBuy beer"), call("\n")],
                         mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--file", "-"])
    @patch.object(sys.stdin, "readline")
    @patch.object(sys.stdout, "write")
    def test_reading_stdin(self, mock_stdout_write, mock_stdin_readline):
        """ Test that tasks can be read from stdin works. """
        mock_stdin_readline.side_effect = ["(B) Call mom\n", "Walk the dog\n", StopIteration]
        next_action()
        self.assertEqual([call("(B) Call mom"), call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--all", "--file", "todo.txt", "--file", "other.txt"])
    @patch("fileinput.open", mock_open(read_data="Call mom\n"))
    @patch.object(sys.stdout, "write")
    def test_reading_two_files(self, mock_stdout_write):
        """ Test that tasks can be read from stdin works. """
        next_action()
        self.assertEqual([call("Call mom [todo.txt]\nCall mom [other.txt]"), call("\n")],
                         mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "--reference", "always"])
    @patch("fileinput.open", mock_open(read_data="\nWalk the dog @home\nBuy beer\n(B) Call mom\n"))
    @patch.object(sys.stdout, "write")
    def test_reference_filename(self, mock_stdout_write):
        """ Test that the printed next actions reference their filename. """
        expected_filename = os.path.expanduser("~/todo.txt")
        next_action()
        self.assertEqual([call("(B) Call mom [{0}]".format(expected_filename)),
                          call("\n")], mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "@home"])
    @patch("fileinput.open", mock_open(read_data="\nWalk the dog @park\n(B) Call mom\n"))
    @patch.object(sys.stdout, "write")
    def test_unknown_context(self, mock_stdout_write):
        """ Test the response when the context is unknown. """
        next_action()
        self.assertEqual([call("Nothing to do! (warning: unknown context: home)"), call("\n")],
                         mock_stdout_write.call_args_list)

    @patch.object(sys, "argv", ["next-action", "+SpringCleaning", "+AutumnCleaning"])
    @patch("fileinput.open", mock_open(read_data="\nWalk the dog @park\n(B) Call mom\n"))
    @patch.object(sys.stdout, "write")
    def test_unknown_project(self, mock_stdout_write):
        """ Test the response when the project is unknown. """
        next_action()
        self.assertEqual(
            [call("Nothing to do! (warning: unknown projects: AutumnCleaning, SpringCleaning)"), call("\n")],
            mock_stdout_write.call_args_list)
